/****************************************************************
*
* Handles low-level CAN bus communication.
*
*****************************************************************/

#include "iodefine.h"
#include "interrupt_handlers.h"
#include "GlobVar.h"
#include "CAN.h"

#include "LED.h"
#include "DebugCom.h"

// Data received over the CAN bus.
unsigned char CAN_data[NUM_DATA_BYTES];
unsigned long CAN_identifier;
unsigned char CAN_sender;
unsigned char CAN_receiver;
unsigned char CAN_datatype;
unsigned char CAN_function;
unsigned char CAN_priority;

/*
 * Initialises CAN bus communication.
 */
void initCAN(unsigned char moduleID)
{
	unsigned short *work;
	unsigned char mailbox;
	unsigned long identifier, used_mailboxes;
	
	// Clears the standby mode for the HCAN module.
	MST.CR2.BIT._HCAN2 = 0;
	
	// Sets master control register. This clears the reset request bit (MCR0).
	HCAN2.MCR.WORD = 0x0000;
	
	// Clears the Reset/Halt/Sleep interrupt flag (cleared by writing 1).
	HCAN2.IRR.BIT.RSTIF = 1;
	
	// Sets the pin functions of the HCAN2 transmit data pin (PB0) and receive data pin (PB1).
	PFC.PBCR1.BIT.PB1MD = 0;
	PFC.PBCR2.WORD |= 0x000F;
	
	// Sets the HCAN2 transmit data pin (PB0) as output.
	PFC.PBIOR.BIT.B0 = 1;
	// Sets the HCAN2 receive data pin (PB1) as input.
	PFC.PBIOR.BIT.B1 = 0;
	
	// Sets the CAN bit rate (see page 506 in hardware manual).
	// System clock =  = 44 236 800 Hz
	// HCAN2 clock = f_CLK = /2 = 22 118 400 Hz
	// Bit rate = f_CLK / (BRP * (1 + TSEG1 + TSEG2))
	// With settings below:
	// Bit rate = 22 118 400 / (10 * (1 + 5 + 4)) = 221 184 bps
	
	// Sets baud rate prescaler (BRP).
	// b15-b8 = 0000 0000 (reserved)
	// b7-b0 = 0000 1001 => 10 system clocks.
	// => 0000 0000 0000 1001 = 0x0009
	HCAN2.BCR0.WORD = 0x0009;
	
	// Sets bit timing parameters.
	// b15-b12 = 0100 => TSEG1 = 5 time quanta.
	// b11 = 0 (reserved)
	// b10-b8 = 011 => TSEG2 = 4 time quanta.
	// b7-b6 = 00 (reserved)
	// b5-b4 = 00 => Bit synchronisation width = 1 time quantum.
	// b3-b1 = 000 (reserved)
	// b0 = 0 => Bit sampling at one point (end of TSEG1).
	// => 0100 0011 0000 0000 = 0x4300
	HCAN2.BCR1.WORD = 0x4300; // Note! Value in Hibot sample code was 0xb401.
	
	// Initialises mailbox. See HCAN2 application note.
	work = (unsigned short *)0xFFFFB100;
	do {
		*work = 0x0000;
		work++;
	} while(work < (unsigned short *) 0xFFFFB4F4);	
	
	// Configures the mailboxes. There are 32 available mailboxes.
	for(mailbox=0; mailbox<32; mailbox++)
	{
		// Checks if the mailbox shall be configured for reception.
		if ( ((RECEPTION_MAILBOXES>>mailbox)&0x01) == 1 )
		{
			// Note that only word access can be used to set message control fields
			// and the LAFM field. See page 487 (note 3) of the SH2Tiny hardware manual.
			
			// Sets mailbox to be used for reception, and also sets the number of
			// data bytes in the mailbox. Note that only word access can be used to
			// set this register.
			//
			// HCAN2.MB[mailbox].CTRLL[16]:
			// Bit 12-15:	CCM[1]=0 | TTE[1]=0 | NMC[1]=0 | ATX[1]=0
			// Bit 8-11:	DART[1]=0 | MBC[3]=0x02
			// Bit 4-7:		PTE[1]=0 | TCT[1]=0 | [2]=0
			// Bit 0-3:		DLC[4]=NUM_DATA_BYTES
			// => CTRLL = 0000 0010 0000 NUM_DATA_BYTES = (0x0200 | NUM_DATA_BYTES)
			//
			HCAN2.MB[mailbox].CTRLL.WORD = (0x0200 | NUM_DATA_BYTES);
			
			// Constructs the identifier field of the mailbox. Only the receiver address
			// part is relevant since the filter mask of the mailbox will be set so that
			// only these bits are used to decide when to receive a CAN messages.
			identifier = IDENTIFIER(0, 0, 0, moduleID, 0);
			// Sets the mailbox identifier.
			setMailboxIdentifier(mailbox, identifier);
		
			// Sets the local acceptance filter mask so that the mailbox only receives
			// messages whose receiver ID mathces the receiver ID of the mailbox
			// identifier. The receiver ID corresponds to bits 7-13 of the identifier.
			// The other bits in the identifier field of a message are in other words
			// not considered when this node decides if a message shall be received
			// or not. The bit setting in the LAFM registers is defined such that
			// 0 = care and 1 = don't care.
			//
			// See also page 512 in the hardware manual of the SH2Tiny.
			// Note that only word access can be used to set these registers.
			//
			// LAFMH[16] = [1]=0 | STDID_LAFM[11]=0x7FF | [2]=0 | EXTID_LAFM bit 16-17[2] = 0x03
			// => LAFMH = 0111 1111 1111 0011 = 0x7FF3
			// LAFML[16] = EXTID_LAFM bit 0-15[16] = 1100 0000 0111 1111 = 0xC07F
			//
			HCAN2.MB[mailbox].LAFMH.WORD = 0x7FF3;
			HCAN2.MB[mailbox].LAFML.WORD = 0xC07F; // EXTID_LAFM[15:0], bits 7-13 are set to 0 (0 = care, 1 = don't care).
		}
		// Checks if the mailbox shall be configured for transmission.
		else if ( ((TRANSMISSION_MAILBOXES>>mailbox)&0x01) == 1 )
		{
			// Sets mailbox to be used for transmission, and also sets the number of
			// data bytes in the mailbox. Note that only word access can be used to
			// set this register.
			//
			// HCAN2.MB[mailbox].CTRLL[16]:
			// Bit 12-15:	CCM[1]=0 | TTE[1]=0 | NMC[1]=0 | ATX[1]=0
			// Bit 8-11:	DART[1]=0 | MBC[3]=0x01
			// Bit 4-7:		PTE[1]=0 | TCT[1]=0 | [2]=0
			// Bit 0-3:		DLC[4]=NUM_DATA_BYTES
			// => CTRLL = 0000 0010 0000 NUM_DATA_BYTES = (0x0200 | NUM_DATA_BYTES)
			//
			HCAN2.MB[mailbox].CTRLL.WORD = (0x0100 | NUM_DATA_BYTES);
		}
		// Disables the mailbox if it is not used.
		else
		{
			// Sets mailbox as inactive by setting MBC bits to 111.
			HCAN2.MB[mailbox].CTRLL.WORD = 0x0700;
		}
	}
	
	// Enables interrupts from the mailboxes used for either reception
	// or transmission. Note that interrupts are enabled by writing 0.
	used_mailboxes = (RECEPTION_MAILBOXES | TRANSMISSION_MAILBOXES);
	HCAN2.MBIMR.WORD.H = ~((unsigned short)(used_mailboxes >> 16));
	HCAN2.MBIMR.WORD.L = ~((unsigned short)(used_mailboxes & 0xFFFF));
	
	// Enables Mailbox Empty Interrupt and Receive Message Interrupt.
	// Note that interrupts are enabled by writing 0.
	// => IMR = 1111 1110 1111 1101 = 0xFEFD
	HCAN2.IMR.WORD &= 0xFEFD;
	
	// Sets interrupt priority for the CAN module.
	INTC.IPRK.BIT._HCAN2 = CAN_INT_PRIORITY;
}

/*
 * Sets the identifier of the specified CAN mailbox.
 */
void setMailboxIdentifier(unsigned char mailbox, unsigned long identifier)
{
	// The 29-bit identifier is contained in the first 29 bits of the identifier
	// argument of this function.
	// The 29-bit identifier of the mailbox is defined from the 11-bit standard
	// identifier (STDID) and the 18-bit extended identifier (EXTID).
	// See page 447 of the SH2Tiny hardware manual for more details.
	// The identifier is defined as STID[10:0] | EXTID[17:0].
	//
	// HCAN2.MB[mailbox].CTRLH[16]:
	// Bit 15:		Not used[1] = 0
	// Bit 4-14:	STID[11] = (identifier >> 18)
	// Bit 3:		RTR[1] = 0
	// Bit 2:		IDE[1] = 1
	// Bit 0-1:		EXTID bit 16-17[2] = ((identifier >> 16) & 0x03)
	
	// Sets the IDE bit and initialises all other bits to zero.
	HCAN2.MB[mailbox].CTRLH.WORD = 0x0004;
	
	// Sets the identifier.
	HCAN2.MB[mailbox].CTRLH.BIT.STID = (identifier >> 18);				// STDID[10:0]
	HCAN2.MB[mailbox].CTRLH.BIT.EXTID = ((identifier >> 16) & 0x03);	// EXTID[17:16]
	HCAN2.MB[mailbox].CTRLM.WORD = (identifier & 0xFFFF);				// EXTID[15:0]
}

/*
 * Sends a CAN message by entering the specified data bytes into an available
 * CAN mailbox and initiates transmission.
 */
void send_CAN_message(unsigned long identifier, unsigned char *CAN_message)
{
	unsigned char mailbox, i;
	unsigned long mailbox_bit, TXPR_temp;
	
	// There are two 16-bit transmit wait registers. TXPR0 is used for
	// mailboxes 1-15 (mailbox 0 is receive only), while TXPR1 is used
	// for mailboxes 16-31.We assemble both these register values into a
	// single temp variable.
	TXPR_temp = ( (unsigned long)(HCAN2.TXPR.WORD.H << 16) | HCAN2.TXPR.WORD.L );
	
	// Initialises the mailbox bit to mailbox 0, i.e. bit number 1.
	mailbox_bit = 1;
	
	// Searches all mailboxes for an available transmission mailbox which does not
	// currently contain an unsent message. There are 32 mailboxes, but only some
	// of these are configured for transmission.
	// Mailbox 0 and 31 are receive only mailboxes.
	for(mailbox=1; mailbox<31; mailbox++)
	{
		// Moves to the next mailbox.
		mailbox_bit = (1 << mailbox);
		
		// Continues to the next mailbox if the current mailbox is
		// not configured for transmission.
		if ( (TRANSMISSION_MAILBOXES & mailbox_bit) == 0)
			continue;
		
		// Checks if the mailbox is available, i.e. verifies that the mailbox does
		// not contain an unsent message.
		if ( (TXPR_temp & mailbox_bit) == 0 )
		{
			// Sets the mailbox identifier.
			setMailboxIdentifier(mailbox, identifier);

			// Copies the data bytes into the mailbox.
			for (i=0; i<NUM_DATA_BYTES; i++)
			{
				HCAN2.MB[mailbox].MSG_DATA[i] = CAN_message[i];
			}
			
			// Starts transmission by setting the transmit wait bit of the mailbox.
			if (mailbox <= 15)
			{
				// Sets the transmit wait bit of the TXPR0 register.
				HCAN2.TXPR.WORD.L |= (((unsigned short)0x01) << mailbox);
			}
			else
			{
				// Sets the transmit wait bit of the TXPR1 register. Bit 0 of the TXPR1
				// register is the transmit wait bit of mailbox 16.
				HCAN2.TXPR.WORD.H |= (((unsigned short)0x01) << (mailbox-16));
			}
			
			// Exits the for loop.
			break;
			
			/*
			// Debug
			SCI3_writeString("CAN MB ");
			SCI3_writeNumber(mailbox);
			SCI3_writeString(": ");
			for (i=0; i<8; i++)
			{
				SCI3_writeString(" Byte");
				SCI3_writeNumber(i);
				SCI3_writeString(": ");
				SCI3_writeNumber(CAN_message[i]);
			}
			SCI3_writeEndOfLine();*/
		}
	}
}

// Transmission interrupt routine for IRR8, which is called each time
// a CAN message is transmitted error-free.
// See page 519 in the SH2Tiny hardware manual.
void INT_HCAN2_SLE1(void)
{
	// Debug
	//shiftLEDs(1);
//	setLEDstate(LED2, LED_GREEN);
//	setLEDstate(LED3, LED_GREEN);
	
	// Clears the transmission end flag for all mailboxes used for
	// transmission (to clear, write 1). Each flag must be cleared to
	// stop this interrupt routine from being called.
	HCAN2.TXACK.WORD.H = (TRANSMISSION_MAILBOXES >> 16);
	HCAN2.TXACK.WORD.L = (TRANSMISSION_MAILBOXES & 0xFFFF);
}

// Reception interrupt routine for IRR1, which is called each time
// a CAN message is received in a mailbox.
// See page 519 in the SH2Tiny hardware manual.
void INT_HCAN2_RM1(void)
{
	unsigned char mailbox, i;
	unsigned long mailbox_bit, RXPR_temp;
	
	// Debug.
//	shiftLEDs(1);
	
	// There are two 16-bit receive complete registers. RXPR0 is used for
	// mailboxes 0-15, while RXPR1 is used for mailboxes 16-31. We assemble
	// both these register values into a single temp variable.
	RXPR_temp = ( (unsigned long)(HCAN2.RXPR.WORD.H << 16) | HCAN2.RXPR.WORD.L );
	
	// Initialises the mailbox bit to mailbox 0, i.e. bit number 1.
	mailbox_bit = 1;
	
	// Checks all mailboxes for new messages. There are 32 mailboxes, but only some
	// of these are configured for reception of messages.
	for(mailbox=0; mailbox<32; mailbox++)
	{
		// Moves to the next mailbox.
		mailbox_bit = (1 << mailbox);
		
		// Continues to the next mailbox if the current mailbox is
		// not configured for reception.
		if ( (RECEPTION_MAILBOXES & mailbox_bit) == 0)
			continue;
		
		// Checks if there is a new message in the mailbox.
		if ( (RXPR_temp & mailbox_bit) > 0 )
		{
			// Clears the receive complete flag (to clear, write 1).
			if (mailbox <= 15)
				HCAN2.RXPR.WORD.L |= mailbox_bit;
			else
				HCAN2.RXPR.WORD.H |= (1 << (mailbox-16));
		
			// Stores the received data.
			for (i=0; i<NUM_DATA_BYTES; i++)
			{
				CAN_data[i] = HCAN2.MB[mailbox].MSG_DATA[i];
			}
			
			// Gets the message identifier.
			// The 29-bit identifier of the mailbox is defined from the 11-bit standard
			// identifier (STDID) and the 18-bit extended identifier (EXTID).
			// See page 447 of the SH2Tiny hardware manual for more details.
			// The identifier is defined as STID[10:0] | EXTID[17:0].
			CAN_identifier = ( ((HCAN2.MB[mailbox].CTRLH.BIT.STID) << 18)
					| ((HCAN2.MB[mailbox].CTRLH.BIT.EXTID) << 16)
					| (HCAN2.MB[mailbox].CTRLM.WORD) );
					
			CAN_sender = GET_SENDER(CAN_identifier);
			//CAN_receiver = GET_RECEIVER(CAN_identifier);
			CAN_datatype = GET_DATA_TYPE(CAN_identifier);
			CAN_function = GET_FUNCTION(CAN_identifier);
			CAN_priority = GET_PRIORITY(CAN_identifier);
			
			// Processes the received CAN message.
			setVariableDebugMessage("\r\nCANDATA: %d",(unsigned char)CAN_data[2]);
			//setVariableDebugMessage("CANDATA: %d",0);
			CAN_processReceivedMessage(CAN_priority, CAN_function, CAN_datatype, CAN_sender, (unsigned char*)CAN_data);
			
			/*
			// Debug
			setLEDstate(LED1, LED_GREEN);
			setLEDstate(LED3, LED_GREEN);

			// Debug
			SCI3_writeString("CAN MB ");SCI3_writeNumber(mailbox);
			SCI3_writeString(": ");
			SCI3_writeString(" SND: ");SCI3_writeNumber(CAN_sender);
			SCI3_writeString(" REC: ");SCI3_writeNumber(CAN_receiver);
			SCI3_writeString(" DAT: ");SCI3_writeNumber(CAN_datatype);
			SCI3_writeString(" FUN: ");SCI3_writeNumber(CAN_function);
			SCI3_writeString(" PRI: ");SCI3_writeNumber(CAN_priority);
	
			for (i=0; i<8; i++)
			{
				SCI3_writeString(" Byte");
				SCI3_writeNumber(i);
				SCI3_writeString(": ");
				SCI3_writeNumber(CAN_data[i]);
			}
			SCI3_writeEndOfLine();
			*/
		}
	}
}
