/****************************************************************
*
* Handles low-level CAN bus communication.
*
*****************************************************************/

#ifndef _CAN_
#define _CAN_

/*
Each CAN message includes a 29-bit identifier field (11 bits for standard
format and 29 bits for extended format). A message is received by a node
on the CAN bus if the identifier of the message matches the identifier
specified for one of the mailboxes of the node. The mailboxes of each node
are configured (using the filter mask) such that only the receiver ID of the
identifier field is considered when deciding if a node should receive a CAN
message or not.

Note that the CAN bus gives priority to the message whose identifier field
has the lowest value. We therefore define the identifier field such that
high priority messages have low identifier values.

The identifier field is defined as follows:

Identifier (29 bits) =	| Priority (3 bits)		| Function (4 bits)		|
						| Data type (8 bits)	| Receiver ID (7 bits)	|
						| Sender ID (7 bits)	|

*/

// The number of data bytes in a CAN message (between 0 and 8 bytes).
#define NUM_DATA_BYTES	0x08

// Bit pattern indicating the mailboxes used to receive messages.
// There are 32 available mailboxes, where mailbox 0 and 31 are receive-only
// mailboxes. Bit x = 1 means that mailbox x is used for reception.
#define RECEPTION_MAILBOXES ((unsigned long)0x7FFF0000) // Mailboxes 16-30 are used for reception.

// Bit pattern indicating the mailboxes used to transmit messages.
// Bit x = 1 means that mailbox x is used for transmission.
#define TRANSMISSION_MAILBOXES ((unsigned long)0x0000FFFE) // Mailboxes 1-15 are used for transmission.

// Extracts the sender ID from the identifier field.
#define GET_SENDER(identifier)		( identifier & 0x7F ) // bits 0-6

// Extracts the receiver ID from the identifier field.
#define GET_RECEIVER(identifier)	( (identifier>>7) & 0x7F ) // bits 7-13

// Extracts the message data type from the identifier field.
#define GET_DATA_TYPE(identifier)	( (identifier>>14) & 0xFF ) // bits 14-21

// Extracts the message function from the identifier field.
#define GET_FUNCTION(identifier)	( (identifier>>22) & 0x0F ) // bits 22-25

// Extracts the message priority from the identifier field.
#define GET_PRIORITY(identifier)	( (identifier>>26) & 0x07 ) // bits 26-28

// Assembles an identifier field from the parameters.
#define IDENTIFIER(priority, function, datatype, receiverID, senderID)	\
		(	( ((unsigned long)priority) << 26 )	\
		|	( ((unsigned long)function) << 22 )	\
		|	( ((unsigned long)datatype) << 14 )	\
		|	( receiverID << 7 )					\
		|	( senderID ) )

/***************************************
* CAN message PRIORITY
****************************************/
#define CAN_MAXIMUM_PRIORITY		0x01
#define CAN_HIGH_PRIORITY			0x02
#define CAN_MEDIUM_PRIORITY			0x03
#define CAN_LOW_PRIORITY			0x04

/***************************************
* CAN message FUNCTION
****************************************/
#define CAN_READ_DATA				0x01
#define CAN_WRITE_DATA				0x02
#define CAN_SUBSCRIBE_TO_DATA		0x03
#define CAN_UNSUBSCRIBE_TO_DATA		0x04

/***************************************
* CAN message DATA TYPE
****************************************/
#define CAN_NO_DATA					0x01
#define CAN_JOINT_STATUS			0x02
#define CAN_ANGLE_AND_ACCELERATION	0x03
#define CAN_STRAIN_GAUGE_DATA		0x04
#define CAN_ANGLE_REFERENCE			0x05
#define CAN_STOP_MOTION				0x06

void initCAN(unsigned char moduleID);
void setMailboxIdentifier(unsigned char mailbox, unsigned long identifier);
void send_CAN_message(unsigned long identifier, unsigned char *CAN_message);

// Interrupt routines.
// KAN FJERNES???
//void SLE1_IRR8(void);
//void RM1_IRR1(void);

#endif
