/****************************************************************
*
* Handles processing of CAN bus data. The functions in this file
* are the interface towards the low-level CAN communication code
* in CAN.c.
*
*****************************************************************/

#include "iodefine.h"
#include "interrupt_handlers.h"
#include "GlobVar.h"
#include "CAN.h"
#include "CAN_interface.h"
#include "DebugCom.h"
#include "ADC.h"
#include "MambaCom.h"
/*#include "HitecSerialInterface.h"
#include "DynamixelSerialInterface.h"
#include "LED.h"
#include "TempSensor.h"
#include "Accelerometer.h"
#include "ForceSensor.h"
#include "WaterSensor.h"*/

// Array of subscriptions which allows other modules (typically the brain module)
// to subscribe to the sensor data retrieved by this module.
struct Subscription subscriptionTable[MAX_NUM_SUBSCRIPTIONS];

/*
 * Initialises CAN bus communication.
 */
void initCANinterface(void)
{
	char i;
	
	// Initialises the subscription counter.
	subscriptionCounter = 0;
	
	// Initialises the subscription table.
	for(i = 0; i < MAX_NUM_SUBSCRIPTIONS; i++)
	{
		subscriptionTable[i].subscriptionActive	= FALSE;
	}
	
	// Updates the ID of this module.
	CAN_updateModuleID();
	
	// Initialises the CAN bus.
	initCAN(moduleID);
}

/*
 * Updates and returns the ID of this module.
 */
unsigned char CAN_updateModuleID(void)
{
	// The ID is updated by reading the setting of the DIP switch on the power
	// supply board, which is connected to DIO0 (PE0 - PE7) of the SH2Tiny.
	
	// Sets the multiplexed pin functions of PE0 - PE7 so that they are IO pins.
	//PFC.PECRL2.WORD = 0x00;
	
	// Sets pins PE0 - PE7 as input pins.
	//PFC.PEIORL.BYTE.L = 0x00;
	
	// Reads the digital input and updates the ID. We invert the byte in order for
	// the upwards position of a switch to correspond to logic 1.
	moduleID = 1; //~(PE.DRL.BYTE.L);
	
	return moduleID;
}

/*
 * Returns the ID of this module without updating the ID.
 */
unsigned char CAN_getModuleID(void)
{
	return moduleID;
}

/*
 * Processes a new message received over the CAN bus.
 */

void CAN_processReceivedMessage(unsigned char priority, unsigned char function, unsigned char datatype, unsigned char sender, unsigned char* data)
{
	// Investigates the function of the CAN message.
	switch(function)
	{
		case CAN_READ_DATA:
			// Replies to the message.
			//CAN_sendReplyMessage(priority, datatype, sender);
			break;
		case CAN_WRITE_DATA:
			// Stores the parameters in the message.
			CAN_processReceivedParameters(datatype, sender, data);
			break;
		case CAN_SUBSCRIBE_TO_DATA:
			// Registers the subscription.
			//CAN_registerSubscriber(datatype, sender, data[0], data[1]);
			break;
		case CAN_UNSUBSCRIBE_TO_DATA:
			// Removes the subscription.
			//CAN_removeSubscriber(datatype, sender);
			break;
		default:
			break;
	}
}


/*
 * Sends a reply to a received CAN message.
 */
 /*
void CAN_sendReplyMessage(unsigned char priority, unsigned char datatype, unsigned char receiver)
{
	// Initialises the data bytes of the reply message.
	unsigned char CAN_reply_data[8];
	
	// Temporary variables used to construct the CAN message.
	int int_tmp;
	int int_array_tmp[6];
	
	// Variable indicating if the reply message shall be sent or not.
	unsigned char skip_sending = FALSE;
	
	// Investigates what data that the reply message shall include.
	switch(datatype)
	{
		// Replies with joint status parameters.
		case CAN_JOINT_STATUS:
			// Adds the alive counter.
			CAN_reply_data[0] = subscriptionCounter;
			
			// Adds the temperature on the power supply card.
			CAN_reply_data[1] = getTemperature_pwr();
			// Adds the temperature on the lid.
			CAN_reply_data[2] = getTemperature_lid();
			
			// Gets the reading of the water sensor inside the casing.
			int_tmp = getWaterSensorValue_casing();
			// Adds the lo-byte of the value.
			CAN_reply_data[3] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[4] = (unsigned char)(int_tmp >> 8);
			
			// Gets the reading of the secondary water sensor.
			int_tmp = getWaterSensorValue_secondary();
			// Adds the lo-byte of the value.
			CAN_reply_data[5] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[6] = (unsigned char)(int_tmp >> 8);
			break;
		
		// Replies with joint angle and linear acceleration.
		case CAN_ANGLE_AND_ACCELERATION:
			// Gets the joint angle and converts it for sending over CAN.
			int_tmp = (int)( (Hitec_getMeasuredAngle()+150)*10 );
			// Adds the lo-byte of the value.
			CAN_reply_data[0] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[1] = (unsigned char)(int_tmp >> 8);
			
			// Gets the linear acceleration in the X direction.
			int_tmp = (int)( (getAccelerationX()+20)*100 );
			// Adds the lo-byte of the value.
			CAN_reply_data[2] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[3] = (unsigned char)(int_tmp >> 8);
			
			// Gets the linear acceleration in the Y direction.
			int_tmp = (int)( (getAccelerationY()+20)*100 );
			// Adds the lo-byte of the value.
			CAN_reply_data[4] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[5] = (unsigned char)(int_tmp >> 8);
			
			// Gets the linear acceleration in the Z direction.
			int_tmp = (int)( (getAccelerationZ()+20)*100 );
			// Adds the lo-byte of the value.
			CAN_reply_data[6] = (unsigned char)(int_tmp & 0xFF);
			// Adds the hi-byte of the value.
			CAN_reply_data[7] = (unsigned char)(int_tmp >> 8);
			break;
		
		// Replies with the 6 strain gauge readings.
		case CAN_STRAIN_GAUGE_DATA:
			// Gets the 6 strain gauge readings (SG1 - SG6). Each reading is a 10 bit value.
			getStrainGaugeValues(int_array_tmp);
			
			// Adds bits 0 - 7 (8 bits) of SG1.
			CAN_reply_data[0] = (unsigned char)(int_array_tmp[0] & 0xFF);
			// Adds bits 8 - 9 (2 bits) of SG1 and bits 0 - 5 (6 bits) of SG2.
			CAN_reply_data[1] = (unsigned char)( ((int_array_tmp[1] & 0x3F)<<2) | (int_array_tmp[0] >> 8) );
			// Adds bits 6 - 9 (4 bits) of SG2 and bits 0 - 3 (4 bits) of SG3.
			CAN_reply_data[2] = (unsigned char)( ((int_array_tmp[2] & 0x0F)<<4) | (int_array_tmp[1] >> 6) );
			// Adds bits 4 - 9 (6 bits) of SG3.
			CAN_reply_data[3] = (unsigned char)(int_array_tmp[2]>>4);
			
			// Adds bits 0 - 7 (8 bits) of SG4.
			CAN_reply_data[4] = (unsigned char)(int_array_tmp[3] & 0xFF);
			// Adds bits 8 - 9 (2 bits) of SG4 and bits 0 - 5 (6 bits) of SG5.
			CAN_reply_data[5] = (unsigned char)( ((int_array_tmp[4] & 0x3F)<<2) | (int_array_tmp[3] >> 8) );
			// Adds bits 6 - 9 (4 bits) of SG5 and bits 0 - 3 (4 bits) of SG6.
			CAN_reply_data[6] = (unsigned char)( ((int_array_tmp[5] & 0x0F)<<4) | (int_array_tmp[4] >> 6) );
			// Adds bits 4 - 9 (6 bits) of SG6.
			CAN_reply_data[7] = (unsigned char)(int_array_tmp[5]>>4);
			break;
		
		default:
			// Skips sending the reply since the data type was invalid.
			skip_sending = TRUE;
			break;
	}
	
	// Sends the reply message.
	if (skip_sending == FALSE)
		CAN_sendMessage(priority, CAN_WRITE_DATA, datatype, receiver, CAN_reply_data);
}
*/

/*
 * Sends a new CAN message.
 */
void CAN_sendMessage(unsigned char priority, unsigned char function, unsigned char datatype, unsigned char receiver, unsigned char* data)
{
	unsigned long CAN_identifier;
	
	// Debug.
	/*
	SCI3_writeEndOfLine();
	SCI3_writeString("SEND CAN! ");
	SCI3_writeString(" Fun: "); SCI3_writeNumber(function);
	SCI3_writeString(" Type: "); SCI3_writeNumber(datatype);
	SCI3_writeString(" Rec: "); SCI3_writeNumber(receiver);
	SCI3_writeString(" Data0: "); SCI3_writeNumber(data[0]);
	SCI3_writeString(" Data1: "); SCI3_writeNumber(data[1]);
	SCI3_writeString(" Data2: "); SCI3_writeNumber(data[2]);
	SCI3_writeEndOfLine();
	*/
	
	// Constructs the identifier field of the message.
	CAN_identifier = IDENTIFIER(priority, function, datatype, receiver, moduleID);
	
	// Sends the CAN message.
	send_CAN_message(CAN_identifier, data);
}

/*
 * Stores new parameters received over the CAN bus.
 */

void CAN_processReceivedParameters(unsigned char datatype, unsigned char sender, unsigned char* data)
{
	int int_tmp;
	double double_tmp;
	
	// Investigates what data that has been received.
	switch(datatype)
	{
		// A new joint reference angle has been received.
		case CAN_ANGLE_AND_ACCELERATION:
			// Gets the CAN version of the received reference angle.
			int_tmp = (int)( (data[1]<<8) | data[0] );
			// Converts the CAN angle into degrees.
			double_tmp = (double)(0.1*int_tmp - 150);
			MambaCom_saveObtainedMambaAngle(double_tmp);
			// Enables servo motion.
			//Hitec_setMotionEnabled(TRUE);
			// Sets the reference angle.
			//Hitec_setReferenceAngle(double_tmp);
			break;
		// A stop motion command has been received.
		case CAN_STOP_MOTION:
			// Disables servo motion.
			//Hitec_setMotionEnabled(FALSE);
			break;
		default:
			break;
	}
}

/*
 * Sends new data to subscribers.
 *
void CAN_manageSubscriptions(void)
{
	char i;
	
	// Updates the subscription counter.
	subscriptionCounter++;
	
	// Scans the subscription table to see if there are subscribers that shall
	// receive new data.
	for(i = 0; i < MAX_NUM_SUBSCRIPTIONS; i++)
	{
		if(subscriptionTable[i].subscriptionActive == TRUE &&
			subscriptionTable[i].nextSendTime == subscriptionCounter)
		{
			// Sends data to the subscriber.
			CAN_sendReplyMessage(subscriptionTable[i].CAN_priority, subscriptionTable[i].datatype, subscriptionTable[i].subscriberID);
			
			// Calculates the next sending time for this subscription.
			subscriptionTable[i].nextSendTime += subscriptionTable[i].sendPeriod;
		}
	}
}
*/

/*
 * Registers a new subscriber to sensor data from this module.
 */
/*
void CAN_registerSubscriber(unsigned char datatype, unsigned char subscriberID, unsigned char CAN_priority, unsigned char sendPeriod)
{
	char i, availableIndex;
	
	// This variable gives the index in the subscription table where the
	// subscription shall be registered. If this variable is negative, then
	// an entry for this subscription was already present in the table.
	availableIndex = -1;
	
	// Scans all entries in the subscription table to see if there is already
	// an entry with the specified data type and subscriber ID. If such an entry
	// is found, then the priority and send period is updated.
	for (i = 0; i < MAX_NUM_SUBSCRIPTIONS; i++)
	{
		if (subscriptionTable[i].datatype == datatype &&
			subscriptionTable[i].subscriberID == subscriberID)
		{
			// An existing entry was found. Updates the subscription data.
			subscriptionTable[i].subscriptionActive	= TRUE;
			subscriptionTable[i].CAN_priority		= CAN_priority;
			subscriptionTable[i].sendPeriod			= sendPeriod;
			subscriptionTable[i].nextSendTime		= subscriptionCounter + 1;
			// Sets this variable to a negative value to indicate that a new
			// entry for the subscription is not needed.
			availableIndex = -1;
			break;
		}
		else if (availableIndex < 0 && subscriptionTable[i].subscriptionActive == FALSE)
		{
			// This is the index where the subscription will be registered if an
			// existing entry for this subscription is not found in the table.
			availableIndex = i;
		}
	}
	
	// Registers the subscription in the first available location in the
	// subscription table if an existing entry was not found.
	if (availableIndex >= 0)
	{
		subscriptionTable[availableIndex].subscriptionActive	= TRUE;
		subscriptionTable[availableIndex].datatype				= datatype;
		subscriptionTable[availableIndex].subscriberID			= subscriberID;
		subscriptionTable[availableIndex].CAN_priority			= CAN_priority;
		subscriptionTable[availableIndex].sendPeriod			= sendPeriod;
		subscriptionTable[availableIndex].nextSendTime			= subscriptionCounter + 1;
	}
}

/*
 * Removes a subscriber.
 *
void CAN_removeSubscriber(unsigned char datatype, unsigned char subscriberID)
{
	char i;
	
	// Finds and deactivates all entries in the subscription table matching the
	// data type and subscriber ID.
	for(i = 0; i < MAX_NUM_SUBSCRIPTIONS; i++)
	{
		if(subscriptionTable[i].subscriptionActive == TRUE &&
			subscriptionTable[i].datatype == datatype &&
			subscriptionTable[i].subscriberID == subscriberID)
		{
			subscriptionTable[i].subscriptionActive	= FALSE;
		}
	}
}

/*
 * Writes subscription data to the serial port for debug purposes.
 *
void writeSubscriptionDataToSerialPort(void)
{
	char i;
	
	SCI3_writeEndOfLine();
	
	// Prints info about all active subscriptions to the serial port.
	for(i = 0; i < MAX_NUM_SUBSCRIPTIONS; i++)
	{
		if(subscriptionTable[i].subscriptionActive == TRUE)
		{
			SCI3_writeString("Sub: "); SCI3_writeNumber(i);
			SCI3_writeString(" Type: "); SCI3_writeNumber(subscriptionTable[i].datatype);
			SCI3_writeString(" ID: "); SCI3_writeNumber(subscriptionTable[i].subscriberID);
			SCI3_writeString(" Pri: "); SCI3_writeNumber(subscriptionTable[i].CAN_priority);
			SCI3_writeString(" Per: "); SCI3_writeNumber(subscriptionTable[i].sendPeriod);
			SCI3_writeString(" Next: "); SCI3_writeNumber(subscriptionTable[i].nextSendTime);
			SCI3_writeEndOfLine();
		}
	}
	
	SCI3_writeEndOfLine();
}
*/
