/****************************************************************
*
* Handles processing of CAN bus data. The functions in this file
* are the interface towards the low-level CAN communication code
* in CAN.c.
*
*****************************************************************/

/****************************************************************

The following lists the variables communicated over the CAN bus:

unsigned char subscriptionCounter: Counter to indicate that the MCU code is running.
-> Sent over the CAN bus as 8 bit values.

unsigned char temp_pwr: Temperature in degrees on power supply card.
unsigned char temp_lid: Temperature in degrees on lid.
-> Sent over the CAN bus as 8 bit values.

int water_casing: ADC value from water detector inside joint casing.
int water_secondary: ADC value from water detector inside extension module.
-> Sent over the CAN bus as 16 bit values.

double accelerationX: Acceleration in m/s^2 in X direction.
double accelerationY: Acceleration in m/s^2 in Y direction.
double accelerationZ: Acceleration in m/s^2 in Z direction.
-> Converted to unsigned int as (acceleration+20)*100.
-> Sent over the CAN bus as 16 bit values.

double measuredJointAngle: The measured angle in degrees of the Hitec servo.
double referenceJointAngle: The reference angle in degrees of the Hitec servo.
-> Converted to unsigned int as (angle+150)*10.
-> Sent over the CAN bus as 16 bit values.

int SGx: ADC measurement of strain gauge bridge x (x = 1,...,6).
-> Sent over the CAN bus as 10 bit values.

unsigned char temp_propmodule: Temperature in degrees in connected propulsion module.
-> Sent over the CAN bus as 8 bit values.

int water_propmodule: ADC value from water detector in connected propulsion module.
-> Sent over the CAN bus as 16 bit values.

*****************************************************************/

/****************************************************************

Each CAN message contains 8 data bytes (64 bits). With reference to the data
variables listed above, the various CAN messages are defined below.
Note that each row of values listed below is added to the array of CAN data bytes
from right to left. The 8 least significant bits of the rightmost value below are
in other words added to CAN data byte 0.

=======================================
 FUNCTION = CAN_READ_DATA
=======================================
DATA TYPE: This field defines what data that shall be read.
CAN data: The data bytes are ignored.

=======================================
 FUNCTION = CAN_WRITE_DATA
=======================================
DATA TYPE = CAN_JOINT_STATUS
CAN data (56 bits) = | water_secondary (16) | water_casing (16) | temp_lid (8) | temp_pwr (8) | subscriptionCounter (8) |

DATA TYPE = CAN_ANGLE_AND_ACCELERATION
CAN data (64 bits) = | accelerationZ (16) | accelerationY (16) | accelerationX (16) | measuredJointAngle (16) |

DATA TYPE = CAN_STRAIN_GAUGE_DATA
CAN data bytes 3 - 0 (30 bits) = | SG3 (10) | SG2 (10) | SG1 (10) |
CAN data bytes 7 - 4 (30 bits) = | SG6 (10) | SG5 (10) | SG4 (10) |

DATA TYPE = CAN_ANGLE_REFERENCE
CAN data (16 bits) = | referenceJointAngle (16) |

=======================================
 FUNCTION = CAN_SUBSCRIBE_TO_DATA
=======================================
DATA TYPE: This field defines what data to subscribe to.
CAN data (16 bits) = | sendPeriod (8) | CAN_priority (8) |

=======================================
 FUNCTION = CAN_UNSUBSCRIBE_TO_DATA
=======================================
DATA TYPE: This field defines what data to unsubscribe to.
CAN data: The data bytes are ignored.

*****************************************************************/

#ifndef _CAN_INTERFACE_
#define _CAN_INTERFACE_

// The maximum number of subscriptions to sensor data.
#define MAX_NUM_SUBSCRIPTIONS	20

// The ID of this module.
unsigned char moduleID;

// The counter used to control when subscribed data is sent.
unsigned char subscriptionCounter;

// Structure holding information about a subscription.
struct Subscription
{
	// The data type of this subscription.
	unsigned char subscriptionActive;
	// The data type of this subscription.
	unsigned char datatype;
	// The CAN ID of the subscriber.
	unsigned char subscriberID;
	// The CAN priority of subscribed data.
	unsigned char CAN_priority;
	// The period at which the subscribed data shall be sent.
	unsigned char sendPeriod;
	// The time at which the next subscription data shall be sent.
	unsigned char nextSendTime;
};

// Function declerations.
void initCANinterface(void);
unsigned char CAN_updateModuleID(void);
unsigned char CAN_getModuleID(void);
void CAN_processReceivedMessage(unsigned char priority, unsigned char function, unsigned char datatype, unsigned char sender, unsigned char* data);
void CAN_sendReplyMessage(unsigned char priority, unsigned char datatype, unsigned char receiver);
void CAN_sendMessage(unsigned char priority, unsigned char function, unsigned char datatype, unsigned char receiver, unsigned char* data);
void CAN_processReceivedParameters(unsigned char datatype, unsigned char sender, unsigned char* data);
void CAN_manageSubscriptions(void);
void CAN_registerSubscriber(unsigned char datatype, unsigned char subscriberID, unsigned char CAN_priority, unsigned char sendPeriod);
void CAN_removeSubscriber(unsigned char datatype, unsigned char subscriberID);
void writeSubscriptionDataToSerialPort(void);

#endif
